# -*- coding: utf-8 -*-

from PDFRenderers import renderers
from utils import normalize
import tempfile, os

class EmptyPDFDocument(Exception):
    pass

class OrderedDict(dict):
    """
    Dictionnary with order
    """
    def __init__(self, *args, **kwargs):
        self.order = []
        dict.__init__(self, *args, **kwargs)

    def __setitem__(self, key, value):
        if not key in self:
            self.order.append(key)
        dict.__setitem__(self, key, value)

    def __delitem__(self, key):
        dict.__delitem__(self, key)
        self.order.remove(key)

    def keys(self):
        return self.order

class PDFDocument(object):
    """
    This a pure Python class, storing a PDF document
    """

    def __init__(self, context = None):
        """
        Constructor
        """
        self.sections = OrderedDict()
        self.config = {}
        self.title = ""
        self.author = ""
        self.context = context
        self.configure()

    def configure(self, cleanup = True,
                  htmltoc = False, htmlnumbertoc = True,
                  htmlpapersize = "a4",
                  papersize = "a4paper", orientation = "portrait",
                  textoc = False,
                  documentclass = "book",
                  level1 = "part*", level2 = "chapter*",
                  pdflatex = "pdflatex",
                  htmldoc = "htmldoc",
                  htmldoc_options = "",
                  extra_headers = ""):
        """
        Configure the generator
        """
        self.config["cleanup"] = cleanup
        self.config["htmltoc"] = htmltoc
        self.config["htmlnumbertoc"] = htmlnumbertoc
        self.config["papersize"] = papersize
        self.config["htmlpapersize"] = htmlpapersize
        self.config["orientation"] = orientation

        self.config["textoc"] = textoc
        self.config["documentclass"] = documentclass
        self.config["level1"] = level1
        self.config["level2"] = level2

        self.config["pdflatex"] = pdflatex
        self.config["htmldoc"] = htmldoc
        self.config["htmldoc_options"] = htmldoc_options

        self.config["extra_headers"] = extra_headers

    def reconfigure(self, **kwargs):
        """
        Change some options
        """
        for key, value in kwargs.items():
            if not key in self.config:
                raise KeyError, key
            self.config[key] = value

    def set_title(self, title = "", author = ""):
        """
        Set the document title
        """
        self.title = title
        self.author = author

    #
    # General adders
    #

    def add_html(self, content, encoding = "utf-8", section = "", title = None, source_object = None, **kwargs):
        """
        Add HTML content to a section (or many)
        Options can be specified to override the main ones
        """
        if not isinstance(content, unicode):
            content = unicode(content, encoding)
        content = content.encode("iso8859-1", "ignore")
        content = content.replace('content="text/html;charset=%s"'  % encoding,
                                  'content="text/html;charset=iso8859-1"')
        self.add_data(content, "text/html", section, title, source_object, **kwargs)

    def add_attachment(self, data, section = "Appendices", title = None, **kwargs):
        """
        Add an attachment, which must itself be a pdf file
        Options can be specified to override the main ones
        """
        self.add_data(data, "application/pdf", section, title, **kwargs)
           
    def add_tex(self, data, section = "", title = None, **kwargs):
        """
        Add TeX data
        """
        self.add_data(data, "text/x-tex", section, title, **kwargs)

    def add_text(self, data, section = "", title = None, **kwargs):
        """
        Add raw text
        """
        self.add_data(data, "text/plain", section, title, **kwargs)

    def add_data(self, data, mimetype, section = "", title = None, source_object = None, **kwargs):
        """
        Add any kind of data to a section
        Options can be specified to override the main ones
        source_object is used only for retrieve relative image in html section
        """
        if not section in self.sections:
            self.sections[section] = []
        if not mimetype in renderers:
            raise NotImplementedError, "no renderer for mime type '%s'"%(mimetype,)
        self.sections[section].append((mimetype, data, title, source_object, kwargs))

    #
    # Specific adders
    #

    def add_list(self, items, escape = True, number = False, prune_empty = False, **kwargs):
        """
        Add a list
        """
        enviro = number and "enumerate" or "itemize"
        res = "\\begin{%s}\n" % enviro
        for item in items:
            if not item and prune_empty:
                continue
            item = normalize(item, escape)
            res += "\\item %s\n" % item
        res += "\\end{%s}\n" % enviro
        self.add_tex(res, **kwargs)

    def add_description(self, items, escape = True, prune_empty = False, **kwargs):
        """
        Add a description list (key, value) pairs
        """
        res = "\\begin{description}\n"
        for key, value in items:
            if not value and prune_empty:
                continue
            key = normalize(key, escape)
            value = normalize(value, escape)
            res += "\\item[%s] %s\n" % (key, value)
        res += "\\end{description}\n"
        self.add_tex(res, **kwargs)

    #
    # Generate the code
    #
            
    def generate(self):
        """
        Generate the PDF document
        """
        tempdir = tempfile.mkdtemp()
        try:
            return self.generate_in(tempdir)
        finally:
            self.cleanup(tempdir)

    def cleanup(self, tempdir):
        """
        Recursively remove temporary directory
        """
        if not self.config["cleanup"]:
            return
        
        for root, dirs, files in os.walk(tempdir, topdown=False):
            for name in files:
                os.remove(os.path.join(root, name))
            for name in dirs:
                os.rmdir(os.path.join(root, name))
        os.rmdir(tempdir)
                    
    def generate_in(self, tmpdir):
        """
        Generate PDF document in current directory
        """
        template = r"""\documentclass[%(papersize)s,%(orientation)s,oneside]{%(documentclass)s}
\usepackage{pdfpages}
\usepackage[utf8]{inputenc}
%(extra_headers)s
\begin{document}
"""
        filename = os.path.join(tmpdir, "main.tex")
        f = open(filename, "w")
        f.write(template % self.config)

        if self.title:
            f.write(r"""\title{%s}
\author{%s}
\maketitle
""" % (self.title, self.author))
        

        if self.config["textoc"]:
            f.write("\\tableofcontents\n")

        partid = 0

        if not self.sections:
            raise EmptyPDFDocument

        for section in self.sections.keys():
            if section:
                f.write("\\%s{%s}\n" % (self.config["level1"], section))
            for mimetype, data, title, source_object, options in self.sections[section]:                
                config = dict(self.config)
                config["tmpdir"] = tmpdir
                config["source_object"] = source_object
                config.update(options)
                renderer = renderers[mimetype]
                partid += 1
                renderer = renderer(config, "part%d" % partid, data, self.context)
                if title:
                    f.write("\\%s{%s}\n" % (self.config["level2"], title))

                f.write(renderer.tex())

        f.write("\\end{document}\n")
        f.close()
        del f

        cmdline = "cd %s; %s --interaction batchmode %s" % (tmpdir, self.config["pdflatex"], filename)
        os.system(cmdline)
        if self.config["textoc"]:
            # For toc, we need to tex it twice
            os.system(cmdline)

        try:
            f = open(os.path.join(tmpdir, "main.pdf"))
        except IOError:
            raise EmptyPDFDocument
        data = f.read()
        f.close()
        del f
        return data
